/*
 
 Copyright (c) 2004 PXI Project Team
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in the
 Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 */

/*

       __   ___  / /
   ___( (\_( ) )/ /
  |\___` \\_\/_/_/_\
  | | ____/\_\`._._...^^^^^^^^^^^^^^^^^^\
  | ||\__/ '_\ \`_._\.;    \_\ |\\)\)\|  \
   \|| | | | | |    | \     \ \|-\\ \ |   \
    || | |^| | |    |  \       __          \
     \ |     | |    |   \ \)\\) \ \_\ |\|\\|\
      \|ejm__| |____|    \ \)\\\ \ \ \|/|-\| \
                         ^^^^^^^^^^^^^^^^^^^^^^

Happy 20th Birthday Tyler Vano    :)     6/8/04

*/


#include <string>
#include <vector>
#include <iostream>
#include "stringmaniplib.h"
#include "binarystream.h"
#include "bnfgrammar.h"
using namespace std;

bool bnf_simplify = true;

string bnf_reformat(const string & s)
{
	// DEBUGGING_CODE! :: cout << "BNF_Reformat Called" << endl;
	string spaceyChars = "[]{}()=|+*";
	string ss = s;
	string t;
	unsigned long int i;
	for (i = ss.size(); i > 0; i--)
	{
		t = mid(ss, i, 1);
		if (t[0] == 32)
		{
			t = mid(ss, i+1, 1);
			if (t[0] == 32)
			{
				ss = left(ss, i) + mid(ss, i+2);
			}
		}
		else if (instr(spaceyChars, t) > 0)
		{
			t = mid(ss, i-1, 1);
			if (t[0] != 92)
			{
				if (t[0] != 32)
				{
					ss = left(ss, i-1) + (string)" " + mid(ss, i);
				}
				t = mid(ss, i+1, 1);
				if (t[0] != 32)
				{
					ss = left(ss, i) + (string)" " + mid(ss, i+1);
				}
			}
		}
	}
	return trim(ss);
}

string bnf_decode_esc_str(const string & s)
{
	// DEBUGGING_CODE! :: cout << "BNF_Decode Called" << endl;
	string ss = s;
	unsigned long int i;
	for (i = 0; i < ss.size(); i++)
	{
		if (ss[i] == 92)
		{
			ss = left(ss, i) + mid(ss, i+2);
		}
	}
	return ss;
}

string bnf_encode_esc_str(const string & s)
{
	// DEBUGGING_CODE! :: cout << "BNF_Encode Called" << endl;
	string escapeyChars = "[]{}()<>=|@\\+*";
	string ss = s;
	string t = " ";
	signed long int i; /* changing this to unsigned has dire consequences!!! believe me, i know!!! */
	for (i = (signed long int)ss.size()-1; i >= 0; i--)
	{
		t[0] = ss[i];
		if (instr(escapeyChars, t) > 0)
		{
			ss = left(ss,i) + (string)"\\" + mid(ss,i+1);
		}
	}
	return ss;
}

void grammarFile::clear(void)
{
	// DEBUGGING_CODE! :: cout << "GrammarFile::Clear Called" << endl;
	if (!rules.empty()) { rules.clear(); }
	script = (string)"";
}

void grammarRule::clear(void)
{
	// DEBUGGING_CODE! :: cout << "GrammarRule::Clear Called" << endl;
	type = 0;
	name = (string)"";
	pragmaDef = (string)"";
	if (!configurations.empty()) { configurations.clear(); }
}

void grammarFactor::clear(void)
{
	// DEBUGGING_CODE! :: cout << "GrammarFactor::Clear Called" << endl;
	type = 0;
	optional = false;
	value = (string)"";
	compilerHandler = (string)"";
	if (!factors.empty()) { factors.clear(); }
}

void grammarFile::readFile(const string & s)
{
	// DEBUGGING_CODE! :: cout << "GrammarFile::ReadFile Called" << endl;
	binarystream b(s, false, false);
	if (b.ok())
	{
		readData(b.readStringValueAsString(b.getLength()));
		b.close();
	}
}

void grammarFile::writeFile(const string & s)
{
	// DEBUGGING_CODE! :: cout << "GrammarFile::WriteFile Called" << endl;
	binarystream b(s, true, true);
	string ss = writeData();
	if (b.ok())
	{
		b.write(ss, ss.size());
		b.close();
	}
}

void grammarFile::readData(const string & data)
{
	// DEBUGGING_CODE! :: cout << "GrammarFile::ReadData Called" << endl;
	int o, o1, o2;
	string s = data;
	string cl;
	int m = 0;
	grammarRule gr;
	string cr = " ";
	string lf = " ";
	cr[0] = 13;
	lf[0] = 10;
	while (s.size() > 0)
	{
		o1 = instr(s, cr);
		o2 = instr(s, lf);
		o1 = (o1 <= 0)?(s.size()+1):o1;
		o2 = (o2 <= 0)?(s.size()+1):o2;
		o = (o1 < o2)?o1:o2;
		cl = trim(left(s, o-1));
		s = mid(s, o+1);
		if (cl.size() > 0)
		{
			if (cl[0] == '.')
			{
				if (cie(cl, (string)".grammar")) { m = 1; }
				else if (cie(cl, (string)".code")) { m = 2; }
				else if (cie(cl, (string)".script")) { m = 2; }
				else if (cie(cl, (string)".compiler")) { m = 2; }
				else if (cie(cl, (string)".language")) { m = 3; }
				else if (cie(cl, (string)".info")) { m = 4; }
				else { m = 0; }
			}
			else if (cie(left(cl, 21), (string)"#pragma languageName ")) { languageName += trim(mid(cl, 22)); }
			else if (cie(left(cl, 20), (string)"#pragma grammarInfo ")) { grammarInfo += (trim(mid(cl, 21)) + cr); }
			else if (cl[0] != '/')
			{
				if (m == 1)
				{
					gr.clear();
					gr.readData(bnf_reformat(cl));
					rules.push_back(gr);
				}
				else if (m == 2)
				{
					script += cl;
					script += cr;
				}
				else if (m == 3)
				{
					languageName += cl;
				}
				else if (m == 4)
				{
					grammarInfo += cl;
					grammarInfo += cr;
				}
			}
		}
	}
	grammarInfo = left(grammarInfo, grammarInfo.size()-1);
}

string grammarFile::writeData(void)
{
	// DEBUGGING_CODE! :: cout << "GrammarFile::WriteData Called" << endl;
	string s;
	vector<grammarRule>::iterator i;
	string cr = " ";
	cr[0] = 13;
	s = ((string)".language" + cr + languageName + cr + cr + (string)".info" + cr + grammarInfo + cr + cr + (string)".grammar" + cr);
	if (!rules.empty())
	{
		for (i = rules.begin(); i < rules.end(); i++)
		{
			s += (i->writeData());
			s += cr;
		}
	}
	s += (cr + (string)".code" + cr + script + cr);
	return s;
}

void grammarRule::readData(const string & s)
{
	// DEBUGGING_CODE! :: cout << "GrammarRule::ReadData Called" << endl;
	int i;
	int m = countfields(s, " ");
	grammarFactor curConfig;
	vector<string> ts;   /* tokenized string */
	vector<string>::iterator ti;   /* token index */
	string ct;   /* current token */
	for (i = 1; i <= m; i++) { ts.push_back(nthfield(s, " ", i)); }

	ti = ts.begin();
	if ( *ti == (string)"#pragma" )
	{
		type = BNF_GR_TYPE_PRAGMA;
		ti++;
		name = *ti;
		ti++;
		while (ti < ts.end())
		{
			pragmaDef += (*ti + (string)" ");
			ti++;
		}
		pragmaDef = trim(pragmaDef);
	}
	else if ( (left(*ti, 1) == (string)"<") && (*(ti+1) == (string)"=") )
	{
		type = BNF_GR_TYPE_DEFINITION;
		name = *ti;
		ti++;
		ti++;
		while (ti < ts.end())
		{
			curConfig.clear();
			curConfig.readData(ts, ti);
			configurations.push_back(curConfig);
			if (ti < ts.end()) { ti++; }   /* skip over pipe that readData ends on */
		}
	}
}

string grammarRule::writeData(void)
{
	// DEBUGGING_CODE! :: cout << "GrammarRule::WriteData Called" << endl;
	string s;
	vector<grammarFactor>::iterator i;
	if (type == BNF_GR_TYPE_PRAGMA)
	{
		return ((string)"#pragma " + name + (string)" " + pragmaDef);
	}
	else if (type == BNF_GR_TYPE_DEFINITION)
	{
		s = name + (string)" = ";
		for (i = configurations.begin(); i < configurations.end(); i++)
		{
			s += (i->writeData());
			if (i < (configurations.end() - 1))
			{
				s += (string)" | ";
			}
		}
		return s;
	}
	return (string)"";
}

void grammarFactor::readData(vector<string> & v, vector<string>::iterator & i)
{
	// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Called" << endl;
	grammarFactor f;
	grammarFactor ff;
	vector<grammarFactor>::iterator ii;
	type = BNF_GF_TYPE_SEQUENCE;
	optional = false;
	duplicity = BNF_DP_TYPE_NORMAL;
	while ( (i < v.end()) && (*i != (string)"]") && (*i != (string)"}") && (*i != (string)"|") )
	{
		f.clear();
		if (*i == (string)"[")
		{
			// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Declaring F Optional" << endl;
			if (i < v.end()) { i++; }
			f.readData(v, i);
			/* f.type = 3; */
			f.optional = true;
			if (i < v.end()) { i++; }
		}
		else if (*i == (string)"{")
		{
			// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Declaring F Choice" << endl;
			f.type = BNF_GF_TYPE_CHOICE;
			f.optional = false;
			if (i < v.end()) { i++; }
			while ( (i < v.end()) && (*i != (string)"}") )
			{
				// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Declaring FF" << endl;
				ff.clear();
				ff.readData(v, i);
				f.factors.push_back(ff);
				if (i < v.end())
				{
					if (*i != (string)"}") { i++; }
				}
			}
			if (i < v.end()) { i++; }
		}
		else
		{
			if ( left(*i, 1) == (string)"<" )
			{
				// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Declaring F RuleRef" << endl;
				f.type = BNF_GF_TYPE_RULE_REF;
				f.value = *i;
			}
			else if ( left(*i, 1) == (string)"@" )
			{
				// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Declaring F BuiltIn" << endl;
				f.type = BNF_GF_TYPE_BUILT_IN;
				f.value = *i;
			}
			else
			{
				// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Declaring F Literal" << endl;
				f.type = BNF_GF_TYPE_LITERAL;
				f.value = bnf_decode_esc_str(*i);
			}
			if (i < v.end()) { i++; }
			if (i < v.end())
			{
				if (*i == (string)"(")
				{
					// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData SettingCompilerHandler" << endl;
					if (i < v.end()) { i++; }
					while ((*i != (string)")") && (i < v.end()))
					{
						f.compilerHandler += (*i + (string)" ");
						if (i < v.end()) { i++; }
					}
					if (i < v.end()) { i++; }
					f.compilerHandler = trim(f.compilerHandler);
				}
			}
		}
		// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Setting Duplicity" << endl;
		f.duplicity = BNF_DP_TYPE_NORMAL;
		if (i < v.end())
		{
			if (*i == (string)"*")
			{
				f.duplicity = BNF_DP_TYPE_MULTIPLE;
				if (i < v.end()) { i++; }
			}
		}
		if (i < v.end())
		{
			if (*i == (string)"+")
			{
				f.duplicity = BNF_DP_TYPE_OPTIONAL;
				if (i < v.end()) { i++; }
			}
		}
		// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Pushing Back" << endl;
		factors.push_back(f);
	}
	if (bnf_simplify && (factors.size() == 1))
	{
		// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Simplifying" << endl;
		f = *factors.begin();
		factors.clear();
		type = f.type;
		optional = (optional || f.optional);
		duplicity = f.duplicity;
		value = f.value;
		compilerHandler = f.compilerHandler;
		if (!f.factors.empty())
		{
			// DEBUGGING_CODE! :: cout << "GrammarFactor::ReadData Transferring Factors" << endl;
			for (ii=f.factors.begin(); ii<f.factors.end(); ii++)
			{
				factors.push_back(*ii);
			}
		}
	}
}

string grammarFactor::writeData(void)
{
	// DEBUGGING_CODE! :: cout << "GrammarFactor::WriteData Called" << endl;
	string s = "";
	vector<grammarFactor>::iterator i;

	if (optional) { s += (string)"[ "; }
	if (type == BNF_GF_TYPE_CHOICE) { s += (string)"{ "; }

	if (type == BNF_GF_TYPE_LITERAL) { s += bnf_encode_esc_str(value); }
	if (type == BNF_GF_TYPE_RULE_REF) { s += value; }
	if (type == BNF_GF_TYPE_BUILT_IN) { s += value; }

	if (! factors.empty())
	{
		for (i = factors.begin(); i < factors.end(); i++)
		{
			s += (i->writeData());
			s += (string)" ";
			if ( (type == 4) && (i < (factors.end()-1)) ) { s += (string)"| "; }
		}
	}

	s = trim(s);
	if (type == BNF_GF_TYPE_CHOICE) { s += (string)" }"; }
	if (optional) { s += (string)" ]"; }
	if (duplicity == BNF_DP_TYPE_ZERO_OR_MORE) { s += (string)" *"; }
	if (duplicity == BNF_DP_TYPE_ONE_OR_MORE) { s += (string)" +"; }
	return s;
}

void grammarFile::dump(void)
{
	// DEBUGGING_CODE! :: cout << "GrammarFile::Dump Called" << endl;
	vector<grammarRule>::iterator i;
	cout << "DUMPING GRAMMARFILE:" << endl;
	cout << "LanguageName: " << languageName << endl;
	cout << "GrammarInfo: " << endl << grammarInfo << endl;
	cout << "Rules:" << endl;
	for (i = rules.begin(); i < rules.end(); i++)
	{
		i->dump();
	}
	cout << "Script:" << endl;
	cout << script << endl;
}

string grammarFile::getPragma( const string & s )
{
	vector<grammarRule>::iterator i;
	for( i = rules.begin(); i != rules.end(); i ++ )
	{
		if( ( ( * i ).type == 1 ) && ( strcomp( ( * i ).name, s ) == 0 ) )
		{
			if( ( * i ).pragmaDef != "" )
			{
				return ( * i ).pragmaDef;
			}
			else
			{
				return "true";
			}
		}
	}
	return "\\";
}

grammarRule & grammarFile::getRule( int rtype, const string & s )
{
	vector<grammarRule>::iterator i;
	for( i = rules.begin(); i != rules.end(); i ++ )
	{
		if( ( ( * i ).type == rtype ) && ( strcomp( ( * i ).name, s ) == 0 ) )
		{
			return ( * i );
		}
	}
	throw "Couldn't find rule.";
}

void grammarRule::dump(void)
{
	// DEBUGGING_CODE! :: cout << "GrammarRule::Dump Called" << endl;
	vector<grammarFactor>::iterator i;
	cout << "  DUMPING GRAMMARRULE:" << endl;
	cout << "  Type: " << stype() << endl;
	cout << "  Name: " << name << endl;
	cout << "  PragmaDef: " << pragmaDef << endl;
	cout << "  Configurations:" << endl;
	for (i = configurations.begin(); i < configurations.end(); i++)
	{
		i->dump(2);
	}
}

void grammarFactor::dump(int indent)
{
	// DEBUGGING_CODE! :: cout << "GrammarFactor::Dump Called" << endl;
	string ind = "  ";
	while (ind.size() < (unsigned long int)(indent*2))
	{
		ind += ind;
	}
	ind = left(ind, (indent*2));
	vector<grammarFactor>::iterator i;
	cout << ind << "DUMPING GRAMMARFACTOR:" << endl;
	cout << ind << "Type: " << stype() << endl;
	cout << ind << "Optional: " << optional << endl;
	cout << ind << "Duplicity: " << dtype() << endl;
	cout << ind << "Value: " << value << endl;
	cout << ind << "CompilerHandler: " << compilerHandler << endl;
	cout << ind << "Factors:" << endl;
	for (i = factors.begin(); i < factors.end(); i++)
	{
		i->dump(indent+1);
	}
}

string grammarRule::stype(void)
{
	if (type == 0) { return (string)"Definition (0)"; }
	if (type == 1) { return (string)"Pragma     (1)"; }
	return (string)"Unknown";
}

string grammarFactor::stype(void)
{
	if (type == 0) { return (string)"Literal           (0)"; }
	if (type == 1) { return (string)"Rule Ref          (1)"; }
	if (type == 2) { return (string)"Built-In          (2)"; }
	if (type == 3) { return (string)"Sequence          (3)"; }
	if (type == 4) { return (string)"Choice            (4)"; }
	return ( (string)"Unknown    (" + str(type) + (string)")" );
}

string grammarFactor::dtype(void)
{
	if (duplicity == 0) { return (string)"Singular     (0)"; }
	if (duplicity == 1) { return (string)"Zero or More (1)"; }
	if (duplicity == 2) { return (string)"One or More  (2)"; }
	return ( (string)"Unknown    (" + str(type) + (string)")" );
}
